import functools
import pandas as pd
import fitgrid


def get_lmer_dfbetas(epochs, factor, **kwargs):
    r"""Fit lmers leaving out factor levels one by one, compute DBETAS.

    Parameters
    ----------
    epochs : Epochs
        Epochs object
    factor : str
        column name of the factor of interest
    **kwargs
        keyword arguments to pass on to ``fitgrid.lmer``, like ``RHS``

    Returns
    -------
    dfbetas : pandas.DataFrame
        dataframe containing DFBETAS values

    Examples
    --------
    Example calculation showing how to pass in model fitting parameters::

        dfbetas = fitgrid.utils.lmer.get_lmer_dfbetas(
            epochs=epochs,
            factor='subject_id',
            RHS='x + (x|a)
        )

    Notes
    -----
    DFBETAS is computed according to the following formula [NieGroPel2012]_:

    .. math::

       DFBETAS_{ij} = \frac{\hat{\gamma}_i - \hat{\gamma}_{i(-j)}}{se\left(\hat{\gamma}_{i(-j)}\right)}

    for parameter :math:`i` and level :math:`j` of ``factor``.


    """

    # get the factor levels
    table = epochs.table.reset_index().set_index(
        [epochs.epoch_id, epochs.time]
    )
    levels = table[factor].unique()

    # produce epochs tables with each level left out
    looo_epochs = (
        fitgrid.epochs_from_dataframe(
            table[table[factor] != level],
            time=epochs.time,
            epoch_id=epochs.epoch_id,
            channels=epochs.channels,
        )
        for level in levels
    )

    # fit lmer on these epochs
    fitter = functools.partial(fitgrid.lmer, **kwargs)
    grids = map(fitter, looo_epochs)
    coefs = (grid.coefs for grid in grids)

    # get coefficient estimates and se from leave one out fits
    looo_coefs = pd.concat(coefs, keys=levels, axis=1)
    looo_estimates = looo_coefs.loc[pd.IndexSlice[:, :, 'Estimate'], :]
    looo_se = looo_coefs.loc[pd.IndexSlice[:, :, 'SE'], :]

    # get coefficient estimates from regular fit (all levels included)
    all_levels_coefs = fitgrid.lmer(epochs, **kwargs).coefs
    all_levels_estimates = all_levels_coefs.loc[
        pd.IndexSlice[:, :, 'Estimate'], :
    ]

    # drop outer level of index for convenience
    for df in (looo_estimates, looo_se, all_levels_estimates):
        df.index = df.index.droplevel(level=-1)

    # (all_levels_estimate - level_excluded_estimate) / level_excluded_se
    dfbetas = all_levels_estimates.sub(looo_estimates, level=1).div(
        looo_se, level=1
    )

    return dfbetas.stack(level=0)



import numpy as np
import pandas as pd

from .epochs import Epochs
from . import defaults


def generate(
    n_epochs=10,
    n_samples=100,
    n_categories=2,
    n_channels=32,
    time=defaults.TIME,
    epoch_id=defaults.EPOCH_ID,
    seed=None,
):
    """Return Epochs object with fake EEG data.

    Parameters
    ----------
    n_epochs : int
        number of epochs per category to be generated
    n_samples : int
        number of samples in a single epochs
    n_categories : int
        number of levels of the categorical variable
    n_channels : int
        number of time series representing EEG channels
    time : str, defaults to defaults.TIME
        time column name
    epoch_id : str, defaults to defaults.EPOCH_ID
        epoch identifier column name
    seed=None : {None, int, array_like}, optional
        Random number generation seed. Default=None lets data
        vary from run to run. Set `seed` to a 32-bit unsigned
        integer to generate the same fake data run to run.
        See numpy.random.RandomState for details.


    Returns
    -------
    epochs : Epochs object
        Epochs object containing simulated EEG data.

    Notes
    -----
    ``n_epochs`` and ``n_categories`` interact in the sense that ``n_epochs``
    epochs are generated for each level of the categorical variable. In other
    words, the true number of epochs in the generated data is equal to
    ``n_epochs`` * ``n_categories``.

    For example, the default ``n_epochs = 10`` and ``n_categories
    = 2`` produces 20 epochs, 10 per category.
    """

    df, channels = _generate(
        n_epochs, n_samples, n_categories, n_channels, time, epoch_id, seed
    )
    return Epochs(df, time=time, epoch_id=epoch_id, channels=channels)


def _generate(
    n_epochs, n_samples, n_categories, n_channels, time, epoch_id, seed=None
):
    """Return Pandas DataFrame with fake EEG data, and a list of channels."""

    if seed is not None:
        np.random.seed(seed)

    total = n_epochs * n_samples * n_categories

    categories = np.array([f'cat{i}' for i in range(n_categories)])

    indices = {
        epoch_id: np.repeat(np.arange(n_epochs * n_categories), n_samples),
        time: np.tile(np.arange(n_samples), n_epochs * n_categories),
    }

    predictors = {
        'categorical': np.tile(np.repeat(categories, n_samples), n_epochs),
        'continuous': np.random.uniform(size=total),
    }

    channels = [f'channel{i}' for i in range(n_channels)]
    eeg = {
        channel: np.random.normal(loc=0, scale=30, size=total)
        for channel in channels
    }

    data = {**indices, **predictors, **eeg}

    df = pd.DataFrame(data).set_index([epoch_id, time]).sort_index()

    return df, channels



#: default epoch identifier
EPOCH_ID = 'Epoch_idx'

#: default time identifier
TIME = 'Time'

#: default generic channel list
CHANNELS = [
    'ch00',
    'ch01',
    'ch02',
    'ch03',
    'ch04',
    'ch05',
    'ch06',
    'ch07',
    'ch08',
    'ch09',
    'ch10',
    'ch11',
    'ch12',
    'ch13',
    'ch14',
    'ch15',
]



import numpy as np
from collections import defaultdict, OrderedDict
import subprocess
import re
import ctypes
import sys
import os
import glob
import warnings

MKL = 'mkl'
OPENBLAS = 'openblas'


def get_index_duplicates_table(df, level):
    """Return a string table of duplicate index values and their locations."""

    assert level in df.index.names

    level_values = df.index.get_level_values(level)
    dupe_mask = level_values.duplicated(keep=False)

    dupes = level_values[dupe_mask]
    dupe_indices = np.flatnonzero(dupe_mask)

    dupe_dict = defaultdict(list)
    for index, value in zip(dupe_indices, dupes):
        dupe_dict[value].append(index)

    # hardcoded padding
    msg = f'Duplicates in index level {level}:\n\n'
    msg += '{0:<20} {1}\n'.format(level, 'Locations')
    for value, locations in sorted(dupe_dict.items()):
        locations = (str(item) for item in locations)
        msg += '{0:<20} {1}\n'.format(value, ', '.join(locations))

    return msg


def get_first_group(groupby):

    first_group_name = list(groupby.groups)[0]
    first_group = groupby.get_group(first_group_name)
    return first_group


def deduplicate_list(lst):
    return list(OrderedDict.fromkeys(lst))


class BLAS:
    def __init__(self, cdll, kind):

        if kind not in (MKL, OPENBLAS):
            raise ValueError(
                f'kind must be {MKL} or {OPENBLAS}, got {kind} instead.'
            )

        self.kind = kind
        self.cdll = cdll

        if kind == MKL:
            self.get_n_threads = cdll.MKL_Get_Max_Threads
            self.set_n_threads = cdll.MKL_Set_Num_Threads
        else:
            self.get_n_threads = cdll.openblas_get_num_threads
            self.set_n_threads = cdll.openblas_set_num_threads

    def __repr__(self):
        if self.kind == MKL:
            kind = 'MKL'
        if self.kind == OPENBLAS:
            kind = 'OpenBLAS'
        n_threads = self.get_n_threads()
        return f'{kind} @ {n_threads} threads'


def get_blas_mac(numpy_module):

    COMMAND = 'otool'
    FLAGS = '-L'
    PATTERN = r'^\t@loader_path/(?P<path>.*{}.*) \(.*\)$'

    NUMPY_PATH = os.path.join(numpy_module.__path__[0], 'core')
    MULTIARRAY_PATH = glob.glob(os.path.join(NUMPY_PATH, 'multiarray.*so'))[0]

    otool_result = subprocess.run(
        args=[COMMAND, FLAGS, MULTIARRAY_PATH],
        check=True,
        stdout=subprocess.PIPE,
        universal_newlines=True,
    )

    output = otool_result.stdout

    if MKL in output:
        kind = MKL
    elif OPENBLAS in output:
        kind = OPENBLAS
    else:
        return None

    pattern = PATTERN.format(kind)
    match = re.search(pattern, output, flags=re.MULTILINE)

    if match:
        rel_path = match.groupdict()['path']
        abs_path = os.path.join(NUMPY_PATH, rel_path)
        cdll = ctypes.CDLL(abs_path)
        return BLAS(cdll, kind)
    else:
        return None


def get_blas_linux(numpy_module):

    COMMAND = 'ldd'
    PATTERN = r'^\t.*{}.* => (?P<path>.*) \(0x.*$'

    NUMPY_PATH = os.path.join(numpy_module.__path__[0], 'core')
    MULTIARRAY_PATH = glob.glob(os.path.join(NUMPY_PATH, 'multiarray.*so'))[0]

    ldd_result = subprocess.run(
        args=[COMMAND, MULTIARRAY_PATH],
        check=True,
        stdout=subprocess.PIPE,
        universal_newlines=True,
    )

    output = ldd_result.stdout

    if MKL in output:
        kind = MKL
    elif OPENBLAS in output:
        kind = OPENBLAS
    else:
        return None

    pattern = PATTERN.format(kind)
    match = re.search(pattern, output, flags=re.MULTILINE)

    if match:
        path = match.groupdict()['path']
        cdll = ctypes.CDLL(path)
        return BLAS(cdll, kind)
    else:
        return None


def get_blas(numpy_module):
    """Return BLAS object or None if neither MKL nor OpenBLAS is found."""

    if sys.platform.startswith('linux'):
        return get_blas_linux(numpy_module)
    elif sys.platform == 'darwin':
        return get_blas_mac(numpy_module)

    warnings.warn(
        f'Searching for BLAS libraries on {sys.platform} is not supported.'
    )


class single_threaded:
    def __init__(self, numpy_module):
        self.blas = get_blas(numpy_module)

    def __enter__(self):
        if self.blas is not None:
            self.old_n_threads = self.blas.get_n_threads()
            self.blas.set_n_threads(1)
        else:
            warnings.warn(
                'No MKL/OpenBLAS found, assuming NumPy is single-threaded.'
            )

    def __exit__(self, *args):
        if self.blas is not None:
            self.blas.set_n_threads(self.old_n_threads)
            if self.blas.get_n_threads() != self.old_n_threads:
                message = (
                    f'Failed to reset {self.blas.kind} '
                    f'to {self.old_n_threads} threads (previous value).'
                )
                raise RuntimeError(message)


def design_matrix_is_constant(df, columns, time):
    """Check that values in columns of df do not change within any epoch.

    See Notes for more details.

    Parameters
    ----------
    df : pandas.DataFrame
        dataframe to check
    columns : list of str
        list of column names to be checked
    time : str
        name of the time column

    Returns
    -------
    result : bool
        True if values in specified columns don't change, False otherwise


    Notes
    -----
    We check that from timepoint to timepoint, each epoch has the same value in
    a given column:

                epoch1
        ==============
             A |    B
        ==============
             1      x
             1      x
             1      x
             1      x
             1      x

                epoch2
        ==============
             A |    B
        ==============
             2      y
             2      y
             2      y
             2      y
             2      y

    This is helpful when performing linear regression on an epochs table where
    the predictors vary with epochs (as they are expected to) but stay constant
    from sample to sample, because we can do our modeling much faster.
    """
    gb = df.groupby(time)
    _, group = next(iter(gb))  # first group

    df_columns_values = df[columns].values
    group_columns_values = group[columns].values

    expected_df_values = np.repeat(group_columns_values, len(gb), axis=0)

    return (df_columns_values == expected_df_values).all()



import numpy as np
import pandas as pd

from .errors import FitGridError
from . import tools


class Epochs:
    """Container class used for storing epochs tables and exposing statsmodels.

    Parameters
    ----------
    epochs_table : pandas DataFrame
        long form dataframe containing epochs with equal indices
    time : str
        time column name
    epoch_id : str
        epoch identifier column name
    channels : list of str
        list of channel names to serve as dependent variables

    Returns
    -------
    epochs : Epochs
        epochs object

    """

    def __init__(self, epochs_table, time, epoch_id, channels):

        # channels must be a list of strings
        if not isinstance(channels, list) or not all(
            isinstance(item, str) for item in channels
        ):
            raise FitGridError('channels should be a list of strings.')

        # all channels must be present as epochs table columns
        missing_channels = set(channels) - set(epochs_table.columns)
        if missing_channels:
            raise FitGridError(
                'channels should all be present in the epochs table, '
                f'the following are missing: {missing_channels}'
            )

        if not isinstance(epochs_table, pd.DataFrame):
            raise FitGridError('epochs_table must be a Pandas DataFrame.')

        # these index columns are required for consistency checks
        for item in (epoch_id, time):
            if item not in epochs_table.index.names:
                raise FitGridError(
                    f'{item} must be a column in the epochs table index.'
                )

        # check no duplicate column names in index and regular columns
        names = list(epochs_table.index.names) + list(epochs_table.columns)
        deduped_names = tools.deduplicate_list(names)
        if deduped_names != names:
            raise FitGridError('Duplicate column names not allowed.')

        # make our own copy so we are immune to modification to original table
        table = (
            epochs_table.copy().reset_index().set_index(epoch_id).sort_index()
        )
        assert table.index.names == [epoch_id]

        snapshots = table.groupby(time)

        # check that snapshots across epochs have equal index by transitivity
        prev_group = None
        for idx, cur_group in snapshots:
            if prev_group is not None:
                if not prev_group.index.equals(cur_group.index):
                    raise FitGridError(
                        f'Snapshot {idx} differs from '
                        f'previous snapshot in {epoch_id} index:\n'
                        f'Current snapshot\'s indices:\n'
                        f'{cur_group.index}\n'
                        f'Previous snapshot\'s indices:\n'
                        f'{prev_group.index}'
                    )
            prev_group = cur_group

        if not prev_group.index.is_unique:
            dupes = tools.get_index_duplicates_table(table, epoch_id)
            raise FitGridError(
                f'Duplicate values in {epoch_id} index not allowed:\n{dupes}'
            )

        # checks passed, set instance variables
        self.time = time
        self.epoch_id = epoch_id
        self.channels = channels
        self.table = table
        self._snapshots = snapshots
        self.epoch_index = tools.get_first_group(snapshots).index.copy()

    def distances(self):
        """Return scaled Euclidean distances of epochs from the "mean" epoch.

        Returns
        -------
        distances : pandas Series or DataFrame
            Series or DataFrame with epoch distances

        Notes
        -----
        Distances are scaled by dividing by the max.
        """

        table = self.table.reset_index().set_index([self.epoch_id, self.time])[
            self.channels
        ]

        n_channels = len(table.columns)
        n_epochs = len(table.index.unique(level=self.epoch_id))
        n_samples = len(table.index.unique(level=self.time))

        assert table.values.size == n_channels * n_epochs * n_samples
        values = table.values.reshape(n_epochs, n_samples, n_channels)

        mean = values.mean(axis=0)
        diff = values - mean

        def l2_norm(data, axis=1):
            return np.sqrt((data * data).sum(axis=axis))

        # first n_samples is axis 1, then n_channels, leaving epochs
        distances_arr = l2_norm(l2_norm(diff))
        distances_arr_scaled = distances_arr / distances_arr.max()
        distances = pd.Series(distances_arr_scaled, index=self.epoch_index)

        return distances

    def plot_averages(self, channels=None, negative_up=True):
        """Plot grand mean averages for each channel, negative up by default.

        Parameters
        ----------
        channels : list of str, optional, defaults to all channels
            list of channel names to plot the averages
        negative_up : bool, optional, default True
            by convention, ERPs are plotted negative voltage up

        Returns
        -------
        fig : matplotlib.figure.Figure
            figure containing plots
        axes : numpy.ndarray of matplotlib.axes.Axes
            axes objects
        """

        if channels is None:
            channels = self.channels

        from . import plots

        data = self._snapshots.mean()
        fig, axes = plots.stripchart(data[channels], negative_up=negative_up)
        return fig, axes



import pandas as pd
import matplotlib.pyplot as plt


def stripchart(data, negative_up=True):

    with plt.rc_context({'font.size': 14}):
        if isinstance(data, pd.Series):
            data = data.unstack()
        _, n = data.shape
        fig, axes = plt.subplots(nrows=n, figsize=(16, n * 3), sharey=True)
        data.plot(subplots=True, ax=axes)

        for ax in axes:
            ax.set(xlabel='')

        plt.tight_layout()
        if negative_up:
            plt.gca().invert_yaxis()

    return fig, axes



from math import ceil
from functools import partial
from multiprocessing import Pool
from contextlib import redirect_stdout
from io import StringIO

import numpy as np
import pandas as pd
from statsmodels.formula.api import ols
from tqdm import tqdm

from .errors import FitGridError
from . import tools
from .fitgrid import FitGrid, LMFitGrid, LMERFitGrid


def validate_LHS(epochs, LHS):

    # must be a list of strings
    if not (
        isinstance(LHS, list) and all(isinstance(item, str) for item in LHS)
    ):
        raise FitGridError('LHS must be a list of strings.')

    # all LHS items must be present in the epochs_table
    missing = set(LHS) - set(epochs.table.columns)
    if missing:
        raise FitGridError(
            'Items in LHS should all be present in the epochs table, '
            f'the following are missing: {missing}'
        )


def validate_RHS(RHS):

    # validate RHS
    if RHS is None:
        raise FitGridError('Specify the RHS argument.')
    if not isinstance(RHS, str):
        raise FitGridError('RHS has to be a string.')


def process_key_and_group(key_and_group, function, channels):
    key, group = key_and_group
    results = {channel: function(group, channel) for channel in channels}
    return pd.Series(results, name=key)


def run_model(epochs, function, channels=None, parallel=False, n_cores=4):
    """Run an arbitrary model on the epochs.

    Parameters
    ----------
    epochs : Epochs
        the epochs object on which the model is to be run
    function : Python function
        function that runs a model, see Notes below for details
    channels : list of str
        list of channels to serve as dependent variables
    parallel : bool, defaults to False
        set to True in order to run in parallel
    n_cores : int, defaults to 4
        number of processes to run in parallel

    Returns
    -------
    grid : FitGrid
        a FitGrid object containing the results

    Notes
    -----
    The function should take two parameters, ``data`` and ``channel``, run
    some model on the data, and return an object containing the results.
    ``data`` will be a snapshot across epochs at a single timepoint,
    containing all channels of interest. ``channel`` is the name of the
    target variable that the function runs the model against (uses it as
    the dependent variable).

    Examples
    --------
    Here's an example of a function that can be passed to ``run_model``::

        def regression(data, channel):
            formula = channel + ' ~ continuous + categorical'
            return ols(formula, data).fit()

    """

    _grid = _run_model(
        epochs, function, channels=channels, parallel=parallel, n_cores=n_cores
    )
    return FitGrid(_grid, epochs.epoch_index, epochs.time)


def _run_model(epochs, function, channels=None, parallel=False, n_cores=4):

    if channels is None:
        channels = epochs.channels

    validate_LHS(epochs, channels)

    groups = tqdm(epochs._snapshots)

    processor = partial(
        process_key_and_group, function=function, channels=channels
    )

    if parallel:
        chunksize = ceil(len(groups) / n_cores)
        with tools.single_threaded(np):
            with Pool(n_cores) as pool:
                results = pool.map(processor, groups, chunksize=chunksize)
    else:
        results = map(processor, groups)

    grid = pd.concat(results, axis=1).T
    grid.index.name = epochs.time
    return grid  # dataframe, not FitGrid


def lm_single(data, channel, RHS, eval_env):
    formula = channel + ' ~ ' + RHS
    return ols(formula, data, eval_env=eval_env).fit()


def lm(epochs, LHS=None, RHS=None, parallel=False, n_cores=4, eval_env=4):
    """Run ordinary least squares linear regression on the epochs.

    Parameters
    ----------
    epochs : Epochs
        epochs object on which regression is to be run
    LHS : list of str, optional, defaults to all channels
        list of channels for the left hand side of the regression formula
    RHS : str
        right hand side of the regression formula
    parallel : bool, defaults to False
        change to True to run in parallel
    n_cores : int, defaults to 4
        number of processes to use for computation
    eval_env : int or patsy.EvalEnvironment, defaults to 4
        environment to use for evaluating patsy formulas, see patsy docs

    Returns
    -------
    grid : LMFitGrid
        LMFitGrid object containing the results of the regression

    """

    if LHS is None:
        LHS = epochs.channels

    validate_LHS(epochs, LHS)
    validate_RHS(RHS)

    function = partial(lm_single, RHS=RHS, eval_env=eval_env)

    _grid = _run_model(
        epochs,
        function=function,
        channels=LHS,
        parallel=parallel,
        n_cores=n_cores,
    )

    return LMFitGrid(_grid, epochs.epoch_index, epochs.time)


def lmer_single(
    data, channel, RHS, family, conf_int, factors, permute, ordered, REML
):
    from pymer4 import Lmer

    model = Lmer(channel + ' ~ ' + RHS, data=data, family=family)
    with redirect_stdout(StringIO()) as captured_stdout:
        model.fit(
            summarize=False,
            conf_int=conf_int,
            factors=factors,
            permute=permute,
            ordered=ordered,
            REML=REML,
        )

    # lmer prints warnings, capture them and attach to the model object
    warning = captured_stdout.getvalue()

    model.has_warning = True if warning else False
    model.warning = warning

    captured_stdout.close()

    del model.data
    del model.design_matrix
    del model.model_obj

    return model


def lmer(
    epochs,
    LHS=None,
    RHS=None,
    family='gaussian',
    conf_int='Wald',
    factors=None,
    permute=None,
    ordered=False,
    REML=True,
    parallel=False,
    n_cores=4,
):
    """Fit lme4 linear mixed model by interfacing with R.

    Parameters
    ----------
    epochs : Epochs
        epochs object on which lmer is to be run
    LHS : list of str, optional, defaults to all channels
        list of channels for the left hand side of the lmer formula
    RHS : str
        right hand side of the lmer formula
    family : str, defaults to 'gaussian'
        distribution link function to use
    conf_int : str, defaults to 'Wald'
    factors : dict, optional
        Keys should be column names in data to treat as factors. Values
        should either be a list containing unique variable levels if
        dummy-coding or polynomial coding is desired. Otherwise values
        should themselves be dictionaries with unique variable levels as
        keys and desired contrast values (as specified in R!) as keys.
    permute : int, defaults to None
        if non-zero, computes parameter significance tests by permuting
        test stastics rather than parametrically. Permutation is done by
        shuffling observations within clusters to respect random effects
        structure of data.
    ordered : bool, defaults to False
        whether factors should be treated as ordered polynomial contrasts;
        this will parameterize a model with K-1 orthogonal polynomial
        regressors beginning with a linear contrast based on the factor
        order provided
    REML : bool, defaults to True
        change to False to use ML estimation
    parallel : bool, defaults to False
        change to True to run in parallel
    n_cores : int, defaults to 4
        number of processes to use for computation

    Returns
    -------
    grid : LMERFitGrid
        LMERFitGrid object containing the results of lmer fitting
    """

    if LHS is None:
        LHS = epochs.channels

    validate_LHS(epochs, LHS)
    validate_RHS(RHS)

    function = partial(
        lmer_single,
        RHS=RHS,
        family=family,
        conf_int=conf_int,
        factors=factors,
        permute=permute,
        ordered=ordered,
        REML=REML,
    )
    _grid = _run_model(
        epochs, function, channels=LHS, parallel=parallel, n_cores=n_cores
    )

    return LMERFitGrid(_grid, epochs.epoch_index, epochs.time)



import warnings
import re
import numpy as np
import pandas as pd
import matplotlib as mpl
from matplotlib import pyplot as plt
import fitgrid

# enforce some common structure for summary dataframes
# scraped out of different fit objects.
INDEX_NAMES = ['Time', 'model', 'beta', 'key']

# each model, beta combination has all these values,
# some are per-beta, some are per-model
KEY_LABELS = [
    '2.5_ci',
    '97.5_ci',
    'AIC',
    'DF',
    'Estimate',
    'P-val',
    'SE',
    'SSresid',
    'T-stat',
    'has_warning',
    'logLike',
    'sigma2',
]

# special treatment for per-model values ... broadcast to all params
PER_MODEL_KEY_LABELS = ['AIC', 'SSresid', 'has_warning', 'logLike', 'sigma2']


def summarize(
    epochs_fg,
    modeler,
    LHS,
    RHS,
    parallel=True,
    n_cores=4,
    save_as=None,
    **kwargs,
):
    """Fit the data with one or more model formulas and return summary information.

    Convenience wrapper, useful for keeping memory use manageable when
    gathering betas and fit measures for a stack of models.


    Parameters
    ----------
    epochs_fg : fitgrid.epochs.Epochs
       as returned by `fitgrid.epochs_from_dataframe()` or
       `fitgrid.from_hdf()`, *NOT* a `pandas.DataFrame`.

    modeler : {'lm', 'lmer'}
       class of model to fit, `lm` for OLS, `lmer` for linear mixed-effects.
       Note: the RHS formula language must match the modeler.

    LHS : list of str
       the data columns to model

    RHS : model formula or list of model formulas to fit
       see the Python package `patsy` docs for `lm` formula langauge
       and the R library `lme4` docs for the `lmer` formula langauge.

    parallel : bool

    n_cores : int
       number of cores to use. See what works, but golden rule if running
       on a shared machine.

    save_as : 2-ple of str, optional
       write the summary dataframe to disk with
       `pd.to_hdf(path, key, format='fixed')`

    **kwargs : key=value arguments passed to the modeler, optional


    Returns
    -------
    summary_df : `pandas.DataFrame`
        indexed by `timestamp`, `model_formula`, `beta`, and `key`,
        where the keys are `ll.l_ci`, `uu.u_ci`, `AIC`, `DF`, `Estimate`,
        `P-val`, `SE`, `T-stat`, `has_warning`, `logLike`.


    Examples
    --------

    >>> lm_formulas = [
        '1 + fixed_a + fixed_b + fixed_a:fixed_b',
        '1 + fixed_a + fixed_b',
        '1 + fixed_a,
        '1 + fixed_b,
        '1',
    ]
    >>> lm_summary_df = fitgrid.utils.summarize(
        epochs_fg,
        'lm',
        LHS=['MiPf', 'MiCe', 'MiPa', 'MiOc'],
        RHS=lmer_formulas,
        parallel=True,
        n_cores=4
    )

    >>> lmer_formulas = [
        '1 + fixed_a + (1 + fixed_a | random_a) + (1 | random_b)',
        '1 + fixed_a + (1 | random_a) + (1 | random_b)',
        '1 + fixed_a + (1 | random_a)',
    ]
    >>> lmer_summary_df = fitgrid.utils.summarize(
        epochs_fg,
        'lmer',
        LHS=['MiPf', 'MiCe', 'MiPa', 'MiOc'],
        RHS=lmer_formulas,
        parallel=True,
        n_cores=12,
        REML=False
    )

    """

    FutureWarning('fitgrid summaries are in early days, subject to change')

    # modicum of guarding
    msg = None
    if isinstance(epochs_fg, pd.DataFrame):
        msg = (
            "Convert dataframe to fitgrid epochs with "
            "fitgrid.epochs_from_dataframe()"
        )
    elif not isinstance(epochs_fg, fitgrid.epochs.Epochs):
        msg = f"epochs_fg must be a fitgrid.Epochs not {type(epochs_fg)}"
    if msg is not None:
        raise TypeError(msg)

    # select modler
    if modeler == 'lm':
        _modeler = fitgrid.lm
        _scraper = _lm_get_summaries_df
    elif modeler == 'lmer':
        _modeler = fitgrid.lmer
        _scraper = _lmer_get_summaries_df
    else:
        raise ValueError("modeler must be 'lm' or 'lmer'")

    # single formula -> singleton list
    if isinstance(RHS, str):
        RHS = [RHS]

    # loop through model formulas fitting and scraping summaries
    summaries = []
    for _rhs in RHS:
        summaries.append(
            _scraper(
                _modeler(
                    epochs_fg,
                    LHS=LHS,
                    RHS=_rhs,
                    parallel=parallel,
                    n_cores=n_cores,
                    **kwargs,
                )
            )
        )

    summary_df = pd.concat(summaries)
    _check_summary_df(summary_df)

    del summaries

    if save_as is not None:
        try:
            fname, group = save_as
            summary_df.to_hdf(fname, group)
        except Exception as fail:
            warnings.warn(
                f"save_as={save_as} failed: {fail}. You can try to "
                "save the returned dataframe with pandas.to_hdf()"
            )

    return summary_df


# ------------------------------------------------------------
# private-ish summary helpers for scraping summary info from fits
# ------------------------------------------------------------
def _check_summary_df(summary_df):
    # order matters
    if not (
        summary_df.index.names == INDEX_NAMES
        and all(summary_df.index.levels[-1] == KEY_LABELS)
    ):
        raise ValueError(
            "uh oh ... fitgrid summary dataframe bug, please post an issue"
        )


def _lm_get_summaries_df(fg_ols, ci_alpha=0.05):
    """scrape fitgrid.LMFitgrid OLS info into a tidy dataframe

    Parameters
    ----------
    fg_ols : fitgrid.LMFitGrid

    ci_alpha : float {.05}
       alpha for confidence interval


    Returns
    -------
    summaries_df : pd.DataFrame
       index.names = [`Time`, `model`, `beta`, `key`]
       columns are the `fg_ols` columns


    Notes
    -----
    The `summaries_df` row and column indexes are munged to match
    fitgrid.lmer._get_summaries_df()

    """

    # grab and tidy the formula RHS
    rhs = fg_ols.tester.model.formula.split('~')[1].strip()
    rhs = re.sub(r"\s+", " ", rhs)

    # fitgrid returns them in the last column of the index
    param_names = fg_ols.params.index.get_level_values(-1).unique()

    # fetch a master copy of the model info
    model_vals = []
    model_key_attrs = [
        ("DF", "df_resid"),
        ("AIC", "aic"),
        ("logLike", 'llf'),
        ("SSresid", 'ssr'),
        ("sigma2", 'mse_resid'),
    ]

    for (key, attr) in model_key_attrs:
        vals = None
        vals = getattr(fg_ols, attr).copy()
        if vals is None:
            raise AttributeError(f"model: {rhs} attribute: {attr}")
        vals['key'] = key
        model_vals.append(vals)

    # statsmodels result wrappers have different versions of llf!
    aics = (-2 * fg_ols.llf) + 2 * (fg_ols.df_model + fg_ols.k_constant)
    if not np.allclose(fg_ols.aic, aics):
        msg = (
            "uh oh ...statsmodels OLS aic and llf calculations have changed."
            " please report an issue to fitgrid"
        )
        raise ValueError(msg)

    # build model has_warnings with False for ols
    warnings = pd.DataFrame(
        np.zeros(model_vals[0].shape).astype('bool'),
        columns=model_vals[0].columns,
        index=model_vals[0].index,
    )
    warnings['key'] = 'has_warning'
    model_vals.append(warnings)

    model_vals = pd.concat(model_vals)

    # constants across the model
    model_vals['model'] = rhs

    # replicate the model info for each beta
    # ... horribly redundant but mighty handy when slicing later
    pmvs = []
    for p in param_names:
        pmv = model_vals.copy()
        # pmv['param'] = p
        pmv['beta'] = p
        pmvs.append(pmv)

    pmvs = pd.concat(pmvs).reset_index().set_index(INDEX_NAMES)

    # lookup the param_name specifc info for this bundle
    summaries = []

    # select model point estimates mapped like so (key, OLS_attribute)
    sv_attrs = [
        ('Estimate', 'params'),  # coefficient value
        ('SE', 'bse'),
        ('P-val', 'pvalues'),
        ('T-stat', 'tvalues'),
    ]

    for idx, (key, attr) in enumerate(sv_attrs):
        attr_vals = getattr(fg_ols, attr).copy()  # ! don't mod the _grid
        if attr_vals is None:
            raise AttributeError(f"not found: {attr}")

        attr_vals.index = attr_vals.index.rename(['Time', 'beta'])
        attr_vals['model'] = rhs
        attr_vals['key'] = key

        # update list of beta bundles
        summaries.append(attr_vals.reset_index().set_index(INDEX_NAMES))

    # special handling for confidence interval
    ci_bounds = [
        f"{bound:.1f}_ci"
        for bound in [100 * (1 + (b * (1 - ci_alpha))) / 2.0 for b in [-1, 1]]
    ]
    cis = fg_ols.conf_int(alpha=ci_alpha)

    cis.index = cis.index.rename(['Time', 'beta', 'key'])
    cis.index = cis.index.set_levels(ci_bounds, 'key')
    cis['model'] = rhs
    summaries.append(cis.reset_index().set_index(INDEX_NAMES))

    summaries_df = pd.concat(summaries)

    # add the parmeter model info
    summaries_df = pd.concat([summaries_df, pmvs]).sort_index().astype(float)

    _check_summary_df(summaries_df)

    return summaries_df


def _lmer_get_summaries_df(fg_lmer):
    """scrape a single model fitgrid.LMERFitGrid into a standard summary format

    Note: some values are fitgrid attributes (via pymer), others are derived

    Parameters
    ----------
    fg_lmer : fitgrid.LMERFitGrid

    """

    def scrape_sigma2(fg_lmer):
        # sigma2 is extracted from fg_lmer.ranef_var ...
        # residuals should be in the last row of ranef_var at each Time
        ranef_var = fg_lmer.ranef_var

        # set the None index names
        assert ranef_var.index.names == ['Time', None, None]
        ranef_var.index.names = ['Time', 'key', 'value']

        assert 'Residual' == ranef_var.index.get_level_values(1).unique()[-1]
        assert all(
            ['Name', 'Var', 'Std']
            == ranef_var.index.get_level_values(2).unique()
        )

        # slice out the Residual Variance at each time point
        # and drop all but the Time indexes to make Time x Chan
        sigma2 = ranef_var.query(
            'key=="Residual" and value=="Var"'
        ).reset_index(['key', 'value'], drop=True)

        return sigma2

    # look these up directly
    pymer_attribs = ['AIC', 'has_warning', 'logLike']

    #  x=lmer_fg caclulate or extract from other attributes
    derived_attribs = {
        # fg_lmer.resid comes from pymer wrapping lme4 function resid(object)
        'SSresid': lambda x: x.resid.groupby('Time').apply(
            lambda y: np.sum(y ** 2)
        ),
        'sigma2': lambda x: scrape_sigma2(x),
    }

    # grab and tidy the formulat RHS from the first grid cell
    rhs = fg_lmer.tester.formula.split('~')[1].strip()
    rhs = re.sub(r"\s+", " ", rhs)

    # coef estimates and stats ... these are 2-D
    summaries_df = fg_lmer.coefs.copy()  # don't mod the original

    summaries_df.index.names = ['Time', 'beta', 'key']
    summaries_df = summaries_df.query("key != 'Sig'")  # drop the stars
    summaries_df.index = summaries_df.index.remove_unused_levels()

    summaries_df.insert(0, 'model', rhs)
    summaries_df.set_index('model', append=True, inplace=True)
    summaries_df.reset_index(['key', 'beta'], inplace=True)

    # scrape AIC and other useful 1-D fit attributes into summaries_df
    for attrib in pymer_attribs + list(derived_attribs.keys()):
        # LOGGER.info(attrib)

        # lookup or calculate model measures
        if attrib in pymer_attribs:
            attrib_df = getattr(fg_lmer, attrib).copy()
        else:
            attrib_df = derived_attribs[attrib](fg_lmer)

        attrib_df.insert(0, 'model', rhs)
        attrib_df.insert(1, 'key', attrib)

        # propagate attributes to each beta ... wasteful but tidy
        # when grouping by beta
        for beta in summaries_df['beta'].unique():
            beta_attrib = attrib_df.copy().set_index('model', append=True)
            beta_attrib.insert(0, 'beta', beta)
            summaries_df = summaries_df.append(beta_attrib)

    summaries_df = (
        summaries_df.reset_index()
        .set_index(INDEX_NAMES)
        .sort_index()
        .astype(float)
    )
    _check_summary_df(summaries_df)

    return summaries_df


def _get_AICs(summary_df):
    """collect AICs, AIC_min deltas, and lmer warnings from summary_df

    Parameters
    ----------

    summary_df : multi-indexed pandas.DataFrame
       as returned by `fitgrid.summary.summarize()`

    Returns
    -------
    aics : multi-indexed pandas pd.DataFrame

    """

    # AIC and lmer warnings are 1 per model, pull from the first
    # model coefficient only, typically (Intercept)
    first_param = summary_df.index.get_level_values('beta').unique()[0]
    AICs = pd.DataFrame(
        (
            summary_df.loc[pd.IndexSlice[:, :, first_param, 'AIC'], :]
            .reset_index(['key', 'beta'], drop=True)
            .stack(0)
        ),
        columns=['AIC'],
    )

    AICs.index.names = AICs.index.names[:-1] + ['channel']
    AICs['min_delta'] = np.inf  # init to float
    AICs.sort_index(inplace=True)

    # calculate AIC_min for the fitted models at each time, channel
    for time in AICs.index.get_level_values('Time').unique():
        for chan in AICs.index.get_level_values('channel').unique():
            slicer = pd.IndexSlice[time, :, chan]
            AICs.loc[slicer, 'min_delta'] = AICs.loc[slicer, 'AIC'] - min(
                AICs.loc[slicer, 'AIC']
            )

    # merge in corresponding column of lmer warnings
    has_warnings = pd.DataFrame(
        summary_df.loc[pd.IndexSlice[:, :, first_param, 'has_warning'], :]
        .reset_index(['key', 'beta'], drop=True)
        .stack(0),
        columns=['has_warning'],
    )

    has_warnings.index.names = has_warnings.index.names[:-1] + ['channel']
    has_warnings.sort_index(inplace=True)
    AICs = AICs.merge(has_warnings, left_index=True, right_index=True)
    FutureWarning('coef AICs are in early days, subject to change')
    return AICs


def plot_betas(
    summary_df,
    LHS,
    alpha=0.05,
    fdr=None,
    figsize=None,
    s=None,
    df_func=None,
    **kwargs,
):

    """Plot model parameter estimates for each data column in LHS

    Parameters
    ----------
    summary_df : pd.DataFrame
       as returned by fitgrid.utils.summary.summarize

    LHS : list of str
       column names of the data fitgrid.fitgrid docs

    alpha : float
       alpha level for false discovery rate correction

    fdr : str {None, 'BY', 'BH'}
        Add markers for FDR adjusted significant :math:`p`-values. BY
        is Benjamini and Yekatuli, BH is Benjamini and Hochberg, None
        supresses the markers.
    df_func : {None, function}
        plot `function(degrees of freedom)`, e.g., `np.log10`,  `lambda x: x`

    s : float
       scatterplot marker size for BH and lmer decorations

    kwargs : dict
       keyword args passed to pyplot.subplots()


    Returns
    -------
    figs : list

    """

    def _do_fdr(_fg_beta):
        # FDR helper
        m = len(_fg_beta)
        pvals = _fg_beta['P-val'].copy().sort_values()
        ks = list()

        if fdr == 'BH':
            # Benjamini & Hochberg ... restricted
            c_m = 1
        if fdr == 'BY':
            # Benjamini & Yekatuli general case
            c_m = np.sum([1 / i for i in range(1, m + 1)])

        for k, p in enumerate(pvals):
            kmcm = k / (m * c_m)
            if p <= kmcm * alpha:
                ks.append(k)

        if len(ks) > 0:
            crit_p = pvals.iloc[max(ks)]
        else:
            crit_p = 0.0

        _fg_beta['sig_fdr'] = _fg_beta['P-val'] < crit_p

        # slice out sig ps for plotting
        sig_ps = _fg_beta.loc[_fg_beta['sig_fdr'], :]
        return sig_ps, crit_p
        # --------------------

    figs = list()

    if isinstance(LHS, str):
        LHS = [LHS]
    assert all([isinstance(col, str) for col in LHS])

    if fdr not in ['BH', 'BY', None]:
        raise ValueError(f"fdr must be 'BH', 'BY' or None")

    # defaults
    if figsize is None:
        figsize = (8, 3)

    betas = summary_df.index.get_level_values('beta').unique()
    for col in LHS:
        # for idx, beta in enumerate(betas):
        for beta in betas:

            # f, ax_beta = plt.subplots(len(betas), ncol=1, **kwargs)
            f, ax_beta = plt.subplots(
                nrows=1, ncols=1, figsize=figsize, **kwargs
            )

            # if len(betas) == 1:
            #    ax_beta = [ax_beta]

            # unstack this beta, column for plotting
            fg_beta = (
                summary_df.loc[pd.IndexSlice[:, :, beta], col]
                .unstack(level='key')
                .reset_index('Time')
            )

            # lmer SEs
            fg_beta['mn+SE'] = (fg_beta['Estimate'] + fg_beta['SE']).astype(
                float
            )
            fg_beta['mn-SE'] = (fg_beta['Estimate'] - fg_beta['SE']).astype(
                float
            )

            fg_beta.plot(
                x='Time',
                y='Estimate',
                # ax=ax_beta[idx],
                ax=ax_beta,
                color='black',
                alpha=0.5,
                label=beta,
            )

            # ax_beta[idx].fill_between(
            ax_beta.fill_between(
                x=fg_beta['Time'],
                y1=fg_beta['mn+SE'],
                y2=fg_beta['mn-SE'],
                alpha=0.2,
                color='black',
            )

            # plot transformed df
            if df_func is not None:
                fg_beta['DF_'] = fg_beta['DF'].apply(lambda x: df_func(x))
                fg_beta.plot(
                    x='Time', y='DF_', ax=ax_beta, label=f"{df_func}(df)"
                )

            if s is not None:
                my_kwargs = {'s': s}
            else:
                my_kwargs = {}

            # mark FDR sig ps
            if fdr is not None:
                # optionally fetch FDR adjusted sig ps
                sig_ps, crit_p = _do_fdr(fg_beta)
                ax_beta.scatter(
                    sig_ps['Time'],
                    sig_ps['Estimate'],
                    color='black',
                    zorder=3,
                    label=f'{fdr} FDR p < crit {crit_p:0.2}',
                    **my_kwargs,
                )

            try:
                # color warnings last to mask sig ps
                warn_ma = np.ma.where(fg_beta['has_warning'] > 0)[0]
                # ax_beta[idx].scatter(
                ax_beta.scatter(
                    fg_beta['Time'].iloc[warn_ma],
                    fg_beta['Estimate'].iloc[warn_ma],
                    color='red',
                    zorder=4,
                    label='model warnings',
                    **my_kwargs,
                )
            except Exception:
                pass

            # ax_beta[idx].axhline(y=0, linestyle='--', color='black')
            # ax_beta[idx].legend()

            ax_beta.axhline(y=0, linestyle='--', color='black')
            ax_beta.legend(loc=(1.0, 0.0))

            # title
            # rhs = fg_lmer.formula[col].unique()[0]
            formula = fg_beta.index.get_level_values('model').unique()[0]
            assert isinstance(formula, str)
            # ax_beta[idx].set_title(f'{col} {beta}: {formula}')
            ax_beta.set_title(f'{col} {beta}: {formula}')

            figs.append(f)

    return figs


def plot_AICmin_deltas(summary_df, figsize=None, gridspec_kw=None, **kwargs):
    r"""plot FitGrid min delta AICs and fitter warnings

    Thresholds of AIC_min delta at 2, 4, 7, 10 are from Burnham &
    Anderson 2004, see Notes.

    Parameters
    ----------
    summary_df : pd.DataFrame
       as returned by fitgrid.utils.summary.summarize

    figsize : 2-ple
       pyplot.figure figure size parameter

    gridspec_kw : dict
       matplotlib.gridspec key : value parameters

    kwargs : dict
       keyword args passed to plt.subplots(...)

    Returns
    -------
    f, axs : matplotlib.pyplot.Figure

    Notes
    -----

       [BurAnd2004]_ p. 270-271. Where :math:`AIC_{min}` is the
       lowest AIC value for "a set of a priori candidate models
       well-supported by the underlying science :math:`g_{i}, i = 1,
       2, ..., R)`",

       .. math:: \Delta_{i} = AIC_{i} - AIC_{min}

       "is the information loss experienced if we are using fitted
       model :math:`g_{i}` rather than the best model, :math:`g_{min}`
       for inference." ...

       "Some simple rules of thumb are often useful in assessing the
       relative merits of models in the set: Models having
       :math:`\Delta_{i} <= 2` have substantial support (evidence), those
       in which :math:`\Delta_{i} 4 <= 7` have considerably less support,
       and models having :math:`\Delta_{i} > 10` have essentially no
       support."


    """

    aics = _get_AICs(summary_df)  # long format
    models = aics.index.unique('model')
    channels = aics.index.unique('channel')

    nrows = len(models)

    if figsize is None:
        figsize = (12, 8)

    f, axs = plt.subplots(
        nrows, 2, **kwargs, figsize=figsize, gridspec_kw=gridspec_kw
    )

    for i, m in enumerate(models):
        # channel traces
        if len(models) == 1:
            traces = axs[0]
            heatmap = axs[1]
        else:
            traces = axs[i, 0]
            heatmap = axs[i, 1]

        traces.set_title(f'aic min delta: {m}', loc='left')
        for c in channels:
            min_deltas = aics.loc[
                pd.IndexSlice[:, m, c], ['min_delta', 'has_warning']
            ].reset_index('Time')
            traces.plot(min_deltas['Time'], min_deltas['min_delta'], label=c)
            warn_ma = np.ma.where(min_deltas['has_warning'] > 0)
            traces.scatter(
                min_deltas['Time'].iloc[warn_ma],
                min_deltas['min_delta'].iloc[warn_ma],
                color='red',
                label=None,
            )

        if i == 0:
            # first channel legend left of the main plot
            traces.legend(loc='upper right', bbox_to_anchor=(-0.2, 1.0))

        aic_min_delta_bounds = [0, 2, 4, 7, 10]
        for y in aic_min_delta_bounds:
            traces.axhline(y=y, color='black', linestyle='dotted')

        # heat map
        _min_deltas = (
            aics.loc[pd.IndexSlice[:, m, :], 'min_delta']
            .reset_index('model', drop=True)
            .unstack('channel')
            .astype(float)
        )

        _has_warnings = (
            aics.loc[pd.IndexSlice[:, m, :], 'has_warning']
            .reset_index('model', drop=True)
            .unstack('channel')
            .astype(bool)
        )

        # colorbrewer 2.0 Blues color blind safe n=5
        # http://colorbrewer2.org/#type=sequential&scheme=Blues&n=5
        pal = ['#eff3ff', '#bdd7e7', '#6baed6', '#3182bd', '#08519c']

        cmap = mpl.colors.ListedColormap(pal)
        # cmap.set_over(color='#fcae91')
        cmap.set_over(color='#08306b')  # darkest from Blues n=7
        cmap.set_under(color='lightgray')

        bounds = aic_min_delta_bounds
        norm = mpl.colors.BoundaryNorm(bounds, cmap.N)
        im = heatmap.pcolormesh(
            _min_deltas.index,
            np.arange(len(_min_deltas.columns) + 1),
            _min_deltas.T,
            cmap=cmap,
            norm=norm,
        )

        # fitter warnings mask
        pal_ma = ['k']
        bounds_ma = [0.5]
        cmap_ma = mpl.colors.ListedColormap(pal_ma)
        cmap_ma.set_over(color='crimson')
        cmap_ma.set_under(alpha=0.0)  # don't show goods
        norm_ma = mpl.colors.BoundaryNorm(bounds_ma, cmap_ma.N)
        heatmap.pcolormesh(
            _has_warnings.index,
            np.arange(len(_has_warnings.columns) + 1),
            _has_warnings.T,
            cmap=cmap_ma,
            norm=norm_ma,
        )
        yloc = mpl.ticker.IndexLocator(base=1, offset=0.5)
        heatmap.yaxis.set_major_locator(yloc)
        heatmap.set_yticklabels(_min_deltas.columns)
        plt.colorbar(im, ax=heatmap, extend='max')

    return f, axs



# to install in development mode, from this directory run
#
#     pip install --user -e .
#
# to install stable package systemwide, as root run
#
#     pip install .
#
# http://python-packaging.readthedocs.io/en/latest/minimal.html

import re
from setuptools import find_packages, setup


# from fitgrid.version import __version__
def get_ver():
    with open("./fitgrid/__init__.py", "r") as stream:
        fg_ver = re.search(
            r".*__version__.*=.*[\"\'](?P<ver_str>\d+\.\d+\.\d+\S*)[\'\"].*",
            stream.read(),
        )

    if fg_ver is None:
        msg = f"""
        fitgrid.__init__.py must have an X.Y.Z semantic version, e.g.,

        __version__ = '0.0.0'
        __version__ = '0.0.0-dev.0.0'

        """
        raise ValueError(msg)
    else:
        return fg_ver['ver_str']


def readme():
    with open('README.md') as f:
        return f.read()


setup(
    name='fitgrid',
    version=get_ver(),
    description='Mass multiple regression manager',
    long_description=readme(),
    long_description_content_type='text/markdown',
    author='Thomas P. Urbach, Andrey Portnoy',
    author_email='turbach@ucsd.edu, aportnoy@ucsd.edu',
    url='https://github.com/kutaslab/fitgrid',
    license='BSD-3-Clause',
    packages=find_packages(exclude=['tests']),
    install_requires=[
        'patsy',
        'statsmodels',
        'matplotlib',
        'scipy',
        'tqdm',
        'tables',
    ],
)



import numpy as np
import pandas as pd
import pickle
from functools import lru_cache
import warnings

from .errors import FitGridError
from . import tools


class FitGrid:
    """Hold rERP fit objects.

    FitGrid should not be created directly, the right way to build it is to
    start with an Epochs object and pass it to a function like `fitgrid.lm`.

    Parameters
    ----------
    _grid : Pandas DataFrame
        Pandas DataFrame of fit objects
    epochs_index : pandas Index
        index containing epoch ids
    time : str
        time column name

    Returns
    -------
    grid: FitGrid
        FitGrid object

    Notes
    -----
    Slicing FitGrids is a little different than slicing Pandas DataFrames. For
    instance, we require that the keys in a list used to slice a FitGrid on
    time or channels be unique. The following Pandas quirk is inherited by
    FitGrids: slicing using a list where some keys are present but some
    are missing from a grid succeeds silently and creates columns with the
    missing keys as names. For example, if you have a grid with columns

        'channel1', 'channel2'

    and you slice:

        grid[:, ['channel1', 'blah']]

    this returns a new grid with columns

        'channel1', 'blah'

    where column 'blah' consists of NaN's. Since version 21.0 of Pandas this
    throws a warning and should in future be replaced with a KeyError.
    """

    def __init__(self, _grid, epoch_index, time):

        # check no duplicate column names
        names = list(_grid.columns)
        deduped_names = tools.deduplicate_list(names)
        if deduped_names != names:
            raise FitGridError('Duplicate column names not allowed.')
        self._grid = _grid
        self.epoch_index = epoch_index
        self.time = time
        self.tester = _grid.iloc[0, 0]
        self.channels = list(_grid.columns)

    def __getitem__(self, slicer):
        """Slice grid on time and channels, return new grid with that shape.

        The intended way to slice a FitGrid is to always slice on both time and
        channels, allowing wildcard colons:

            `grid[:, ['channel1', 'channel2']]`

        or

            `grid[:, :]`

        or

            `grid[25:-25, 'channel2']
        """

        if (
            not isinstance(slicer, tuple)
            or not hasattr(slicer, '__len__')
            or len(slicer) != 2
        ):
            raise FitGridError('Must slice on time and channels.')

        # now we can unpack
        time, channels = slicer

        def check_slicer_component(component):
            if isinstance(component, slice):
                return component
            elif isinstance(component, list):
                # deduplicate and warn if duplicates found
                deduped_component = tools.deduplicate_list(component)
                if deduped_component != component:
                    msg = f'Slicer {component} contained duplicates, '
                    msg += f'slicing instead on deduped: {deduped_component}.'
                    warnings.warn(UserWarning(msg))
                return deduped_component
            else:
                # wrap in list to always get a DataFrame in return on slicing
                # otherwise we might get a scalar or a pandas Series,
                # which can't be used to create a FitGrid object
                return [component]

        time = check_slicer_component(time)
        channels = check_slicer_component(channels)
        subgrid = self._grid.loc[time, channels].copy()
        return self.__class__(subgrid, self.epoch_index, self.time)

    @lru_cache()
    def __getattr__(self, name):
        """Broadcast attribute extraction in the grid.

        This is the method that gets called when an attribute is requested that
        FitGrid does not have. We then go and check if the tester has it, and
        broadcast if it does.
        """

        if not hasattr(self.tester, name):
            raise AttributeError(f'No such attribute: {name}.')

        temp = self._grid.applymap(lambda x: getattr(x, name))
        return self._expand(temp)

    def __call__(self, *args, **kwargs):
        """Broadcast method calling in the grid.

        This is the method that gets called when the grid is called as if it
        were a method.
        """
        if not callable(self.tester):
            raise FitGridError(
                f'This grid is not callable, '
                f'current type is {type(self.tester)}'
            )

        # if we are not callable, we'll get an appropriate exception
        temp = self._grid.applymap(lambda x: x(*args, **kwargs))
        return self._expand(temp)

    def __dir__(self):

        # the result of this call is what shows up for tab completion

        # so we add the cell attributes:
        cell_attrs = [
            item for item in dir(self.tester) if not item.startswith('__')
        ]

        # and the attributes of the grid itself:
        grid_attrs = [self.save.__name__]

        return cell_attrs + grid_attrs

    def __repr__(self):

        samples, chans = self._grid.shape
        classname = self.__class__.__name__
        return f'{samples} by {chans} {classname} of type {type(self.tester)}.'

    def save(self, filename):
        """Save FitGrid object to file (reload with ``fitgrid.load_grid``).

        Parameters
        ----------
        filename : str
            file name to use

        """

        with open(filename, 'wb') as file:
            kernel = self._grid, self.epoch_index, self.time
            pickle.dump(kernel, file, protocol=pickle.HIGHEST_PROTOCOL)

    def expand_series_or_df(self, temp):
        """Expand a DataFrame that has Series or DataFrames for values."""

        columns = (
            pd.concat(temp[channel].tolist(), keys=temp.index)
            for channel in temp
        )
        # concatenate columns, channel names are top level columns indices
        result = pd.concat(columns, axis=1, keys=temp.columns)

        # stack to achieve long form if columns have multiple levels
        if isinstance(result.columns, pd.core.indexes.multi.MultiIndex):
            return result.stack()
        return result

    def _expand(self, temp):
        """Expand the values in the grid if possible, return frame or grid."""

        tester = temp.iloc[0, 0]

        # no processing needed
        if np.isscalar(tester):
            return temp

        # familiar types, expand them
        if isinstance(tester, pd.Series) or isinstance(tester, pd.DataFrame):
            # want single index level
            # can get more if original DataFrame had a multiindex
            # in Epochs we ensure that only epoch_id is in the index for
            # groupby
            if tester.index.nlevels > 1:
                raise NotImplementedError(
                    f'index should have one level, have {tester.index.nlevels}'
                    f' instead: {tester.index.names}'
                )
            return self.expand_series_or_df(temp)

        # array-like, try converting to array and then Series/DataFrame
        if isinstance(tester, tuple) or isinstance(tester, list):
            array_form = np.array(tester)

            if array_form.ndim == 1:
                # Tidy 1-D are Series. Untidy 1-D may be broadcastable
                # into a tidy DataFrame

                def series_fun(x):
                    return pd.Series(np.array(x))  # default

                if array_form.dtype != np.dtype("object"):
                    pd_fun = series_fun  # tidy
                else:
                    try:
                        # broadcastable to DataFrame?
                        def df_fun(x):
                            return pd.DataFrame(np.broadcast(*x)).T

                        pd_fun = df_fun
                        pd_fun(tester)
                    except Exception:
                        # oh well, fall back to Series
                        pd_fun = series_fun

                temp = temp.applymap(lambda x: pd_fun(x))

            elif array_form.ndim == 2:
                temp = temp.applymap(lambda x: pd.DataFrame(np.array(x)))
            else:
                raise NotImplementedError(
                    'Cannot use elements with dim > 2,'
                    f'element has ndim = {array_form.ndim}.'
                )
            temp_expanded = self.expand_series_or_df(temp)
            temp_epoch_index = self.add_epoch_index(temp_expanded)
            return temp_epoch_index

        # array, try converting to Series/DataFrame
        if isinstance(tester, np.ndarray):
            if tester.ndim == 1:
                temp = temp.applymap(lambda x: pd.Series(x))
            elif tester.ndim == 2:
                temp = temp.applymap(lambda x: pd.DataFrame(x))
            else:
                raise NotImplementedError(
                    'Cannot use elements with dim > 2,'
                    f'element has ndim = {tester.ndim}.'
                )
            temp_expanded = self.expand_series_or_df(temp)
            temp_epoch_index = self.add_epoch_index(temp_expanded)
            return temp_epoch_index

        # catchall for all types we don't handle explicitly
        # statsmodels objects, dicts, methods all go here
        return FitGrid(temp, self.epoch_index, self.time)

    def add_epoch_index(self, temp):
        """We assume that temp is in long form, the columns are channels, and the
        first index level is time."""

        # first index level is time
        assert temp.index.names[0] == self.time

        # temp should be long form, columns have single level (channels
        # hopefully)
        assert not isinstance(temp.columns, pd.core.index.MultiIndex)

        # we can only handle 2- or 3-dimensional
        assert temp.index.nlevels in (2, 3)

        for i in range(1, temp.index.nlevels):
            level = temp.index.levels[i]
            # if a level looks like it was automatically created by Pandas,
            # we replace it with the epoch_index
            if (
                isinstance(level, pd.RangeIndex)
                and len(level) == len(self.epoch_index)
                and level._start == 0
                and level._step == 1
                and level._stop == len(self.epoch_index)
            ):
                temp.index.set_levels(self.epoch_index, level=i, inplace=True)
                temp.index.rename(self.epoch_index.name, level=i, inplace=True)

        return temp


class LMFitGrid(FitGrid):
    def __dir__(self):

        lmfitgrid_attrs = [
            self.plot_betas.__name__,
            self.plot_adj_rsquared.__name__,
            self.influential_epochs.__name__,
        ]
        return super().__dir__() + lmfitgrid_attrs

    def plot_betas(self, legend_on_bottom=False):
        """Plot betas of the model, one plot per channel, overplotting betas.

        Parameters
        ----------
        legend_on_bottom : bool, defaults to False
            set to True to plot single legend below all channel plots

        Returns
        -------
        fig : matplotlib.figure.Figure
            figure containing plots
        axes : numpy.ndarray of matplotlib.axes.Axes
            axes objects
        """

        import matplotlib.pyplot as plt

        with plt.rc_context({'font.size': 14}):
            params = self.params.unstack()
            channels, betas = params.columns.levels
            figsize = (16, 4 * len(channels))

            n_plots = len(channels) + 1 if legend_on_bottom else len(channels)

            fig, axes = plt.subplots(
                nrows=n_plots, figsize=figsize, sharey=True
            )

            # wrap in list if single axis to allow for zipping later
            if not isinstance(axes, np.ndarray):
                axes = [axes]

            if legend_on_bottom:
                legend_ax = axes[-1]
                channel_axes = axes[:-1]
            else:
                channel_axes = axes

            for ax, chan in zip(channel_axes, channels):
                for beta in params[chan]:
                    ax.plot(params[chan][beta], label=beta)
                ax.set(ylabel=chan, xlabel=params.index.name)
                if not legend_on_bottom:
                    ax.legend(
                        loc='upper center',
                        ncol=len(params[chan]),
                        bbox_to_anchor=(0.5, 1.2),
                        fancybox=True,
                        shadow=False,
                    )

            if legend_on_bottom:
                handles, labels = ax.get_legend_handles_labels()
                legend_ax.set_axis_off()
                legend_ax.legend(handles, labels, mode='expand', ncol=3)

            fig.tight_layout()

            return fig, axes

    def plot_adj_rsquared(self):
        """Plot adjusted :math:`R^2` as a heatmap with marginal bar and line.

        Returns
        -------
        fig : matplotlib.figure.Figure
            figure containing plots
        gs : matplotlib.gridspec.GridSpec
            grid specification that determines locations and sizes of subplots
        bar, heatmap, colorbar, line : matplotlib.axes._subplots.AxesSubplot
            subplot objects
        """

        import matplotlib.pyplot as plt

        rsq_adj = self.rsquared_adj

        with plt.rc_context({'font.size': 14}):
            fig = plt.figure(figsize=(16, 12))
            gs = plt.GridSpec(2, 2, width_ratios=[13, 3], height_ratios=[7, 2])

            bar = plt.subplot(gs[1])
            bar.barh(self._grid.columns, rsq_adj.mean(axis=0))

            heatmap = plt.subplot(gs[0], sharey=bar)
            heatmap_image = heatmap.imshow(rsq_adj.T, aspect='auto')
            heatmap.get_xaxis().set_visible(False)

            colorbar = plt.subplot(gs[3])
            plt.colorbar(mappable=heatmap_image, cax=colorbar, aspect=0.75)

            line = plt.subplot(gs[2])
            line.plot(rsq_adj.mean(axis=1))

            plt.tight_layout()
            plt.margins(x=0)

        return fig, gs, bar, heatmap, colorbar, line

    def influential_epochs(self, top=None):
        """Return dataframe with top influential epochs ranked by Cook's-D.

        Parameters
        ----------
        top : int, optional, default None
            how many top epochs to return, all epochs by default

        Returns
        -------
        top_epochs : pandas DataFrame
            dataframe with epoch_id as index and aggregated Cook's-D as values

        Notes
        -----
        Cook's distance is aggregated by simple averaging across time and
        channels.
        """

        return (
            self.get_influence()
            .cooks_distance.drop(axis=0, index=1, level=1)
            .reset_index(level=1, drop=True)
            .mean(axis=0, level=1)
            .mean(axis=1)
            .sort_values(ascending=False)
            .to_frame(name='average_Cooks_D')
            .iloc[:top]
        )


class LMERFitGrid(FitGrid):
    def __or__(self, other):

        if not isinstance(other, self.__class__):
            raise FitGridError(
                'Can only compare LMERFitGrid to other LMERFitGrid.'
            )
        # figure out the situation
        this_fixef = self.tester.fixef.columns
        other_fixef = other.tester.fixef.columns
        same_fixef = this_fixef.equals(other_fixef)
        at_least_one_fit_with_REML = self.tester._REML or other.tester._REML

        if (not same_fixef) and at_least_one_fit_with_REML:
            raise FitGridError(
                'Cannot compare models with different fixed effects when '
                'REML is used. For more context, see '
                'https://stats.stackexchange.com/a/116796'
            )
        diff = self.AIC - other.AIC

        import matplotlib.pyplot as plt

        with plt.rc_context({'font.size': 14}):
            fig = plt.figure(figsize=(16, 12))
            gs = plt.GridSpec(1, 2, width_ratios=[15, 1])

            heatmap = plt.subplot(gs[0])
            heatmap_image = heatmap.imshow(diff.T, aspect='auto')

            colorbar = plt.subplot(gs[1])
            plt.colorbar(mappable=heatmap_image, cax=colorbar)

        return fig, gs, heatmap, colorbar



import pandas as pd
import pickle
import statsmodels
from statsmodels.regression.linear_model import (
    RegressionResults,
    RegressionResultsWrapper,
)
from .epochs import Epochs
from .fitgrid import FitGrid, LMFitGrid, LMERFitGrid
from .errors import FitGridError
from . import defaults


def epochs_from_hdf(filename, key, time, epoch_id, channels):
    """Construct Epochs object from an HDF5 file containing an epochs table.

    The HDF5 file should contain columns with names defined by `epoch_id` and
    `time` either as index columns or as regular columns. This is added as a
    convenience, in general, input epochs tables should contain these columns
    in the index.

    Parameters
    ----------
    filename : str
        HDF5 file name
    key : str
        group identifier for the dataset when HDF5 file contains more than one
    time : str
        time column name
    epoch_id : str
        epoch identifier column name
    channels : list of str
        list of string channel names

    Returns
    -------
    epochs : Epochs
        an Epochs object with the data
    """

    df = pd.read_hdf(filename, key=key)

    # time and epoch id already present in index
    if epoch_id in df.index.names and time in df.index.names:
        return Epochs(df, time=time, epoch_id=epoch_id, channels=channels)

    # time and epoch id present in columns, set index
    if epoch_id in df.columns and time in df.columns:
        df.set_index([epoch_id, time], inplace=True)
        return Epochs(df, time=time, epoch_id=epoch_id, channels=channels)

    raise FitGridError(
        f'Dataset has to contain {epoch_id} and {time} as columns or indices.'
    )


def epochs_from_dataframe(dataframe, time, epoch_id, channels):
    """Construct Epochs object from a Pandas DataFrame epochs table.

    The DataFrame should contain columns with names defined by epoch_id and
    time as index columns.

    Parameters
    ----------
    dataframe : pandas DataFrame
        a pandas DataFrame object
    time : str
        time column name
    epoch_id : str
        epoch identifier column name
    channels : list of str
        list of string channel names

    Returns
    -------
    epochs : Epochs
        an Epochs object with the data
    """
    return Epochs(dataframe, time=time, epoch_id=epoch_id, channels=channels)


def epochs_from_feather(filename, time, epoch_id, channels):
    """Construct Epochs object from a Feather file containing an epochs table.

    The file should contain columns with names defined by epoch_id and time.

    Parameters
    ----------
    filename : str
        Feather file name
    time : str
        time column name
    epoch_id : str
        epoch identifier column name
    channels : list of str
        list of string channel names

    Returns
    -------
    epochs : Epochs
        an Epochs object with the data
    """

    df = pd.read_feather(filename)

    # time and epoch id present in columns, set index
    if epoch_id in df.columns and time in df.columns:
        df.set_index([epoch_id, time], inplace=True)
        return Epochs(df, time=time, epoch_id=epoch_id, channels=channels)

    raise FitGridError(
        f'Dataset has to contain {epoch_id} and {time} as columns or indices.'
    )


def load_grid(filename):
    """Load a FitGrid object from file (created by running grid.save).

    Parameters
    ----------
    filename : str
        indicates file to load from

    Returns
    -------
    grid : FitGrid
        loaded FitGrid object
    """

    from pymer4 import Lmer

    with open(filename, 'rb') as file:
        _grid, epoch_index, time = pickle.load(file)

    tester = _grid.iloc[0, 0]

    if isinstance(tester, (RegressionResults, RegressionResultsWrapper)):
        return LMFitGrid(_grid, epoch_index, time)
    elif isinstance(tester, Lmer):
        return LMERFitGrid(_grid, epoch_index, time)
    else:
        return FitGrid(_grid, epoch_index, time)



from .fake_data import generate
from .io import (
    epochs_from_hdf,
    epochs_from_dataframe,
    load_grid,
    epochs_from_feather,
)
from .models import run_model, lm, lmer
from . import utils, defaults

__version__ = "0.4.2"



import warnings
import numpy as np
import pandas as pd
import patsy
from tqdm import tqdm
from statsmodels.stats.outliers_influence import (
    variance_inflation_factor as vif,
)
from statsmodels.regression.linear_model import RegressionResultsWrapper
from fitgrid.fitgrid import FitGrid

# ------------------------------------------------------------
# fitgrid's database of statsmodels OLSInfluence diagnostics:
#
#  * what there is, how it runs, and data type, like so
#
#  *  attr : (calc_type, value_dtype, df.index.names)
#
#  df.index.names are as returned by LMFitGrid attribute getter
#  nobs = number of observations
#  nobs_k = number of observations x model regressors
#  nobs_loop = nobs re-fitting ... slow
#  TPU 03/19
# ------------------------------------------------------------
_FLOAT_TYPE = np.float64
_INT_TYPE = np.int64
_OLS_INFLUENCE_ATTRS = {
    '_get_drop_vari': (None, None, None),
    '_res_looo': (None, None, None),
    '_ols_xnoti': (None, None, None),
    'aux_regression_exog': (None, None, None),
    'aux_regression_endog': (None, None, None),
    'cooks_distance': ('nobs', _FLOAT_TYPE, ['Time', None, 'Epoch_idx']),
    'cov_ratio': ('nobs_loop', _FLOAT_TYPE, ['Time', 'Epoch_idx']),
    'det_cov_params_not_obsi': (
        'nobs_loop',
        _FLOAT_TYPE,
        ['Time', 'Epoch_idx'],
    ),
    'dfbetas': ('nobs_loop', _FLOAT_TYPE, ['Time', 'Epoch_idx', None]),
    'dffits': ('nobs_loop', _FLOAT_TYPE, ['Time', None, 'Epoch_idx']),
    'dffits_internal': ('nobs', _FLOAT_TYPE, ['Time', None, 'Epoch_idx']),
    'endog': (None, None, None),  # ('nobs', _FLOAT_TYPE),  # from data
    'ess_press': ('nobs', _FLOAT_TYPE, ['Time']),
    'exog': (None, None, None),  # 'nobs_k', _FLOAT_TYPE),  # from data
    'get_resid_studentized_external': (None, None, None),  # method
    'hat_diag_factor': ('nobs', _FLOAT_TYPE, ['Time', 'Epoch_idx']),
    'hat_matrix_diag': ('nobs', _FLOAT_TYPE, ['Time', 'Epoch_idx']),
    'influence': ('nobs', _FLOAT_TYPE, ['Time', 'Epoch_idx']),
    'k_vars': ('nobs', _INT_TYPE, ['Time']),
    'model_class': (None, None, None),  # not a DataFrame
    'nobs': ('nobs', _INT_TYPE, ['Time']),
    'params_not_obsi': ('nobs_loop', _FLOAT_TYPE, ['Time', 'Epoch_idx', None]),
    'resid_press': ('nobs', _FLOAT_TYPE, ['Time', 'Epoch_idx']),
    'resid_std': ('nobs', _FLOAT_TYPE, ['Time', 'Epoch_idx']),
    'resid_studentized_external': (
        'nobs_loop',
        _FLOAT_TYPE,
        ['Time', 'Epoch_idx'],
    ),
    'resid_studentized_internal': ('nobs', _FLOAT_TYPE, ['Time', 'Epoch_idx']),
    'resid_var': ('nobs', _FLOAT_TYPE, ['Time', 'Epoch_idx']),
    'results': (None, None, None),  # not a DataFrame
    'save': (None, None, None),  # not a DataFrame
    'sigma2_not_obsi': ('nobs_loop', _FLOAT_TYPE, ['Time', 'Epoch_idx']),
    'sigma_est': ('nobs', _FLOAT_TYPE, ['Time']),
    'summary_frame': (None, None, None),  # not a DataFrame
    'summary_table': (None, None, None),  # not a DataFrame
}


def get_vifs(epochs, RHS):
    def get_single_vif(group, RHS):
        dmatrix = patsy.dmatrix(formula_like=RHS, data=group)
        vifs = {
            name: vif(dmatrix, index)
            for name, index in dmatrix.design_info.column_name_indexes.items()
        }
        return pd.Series(vifs)

    tqdm.pandas(desc="Time")

    return epochs._snapshots.progress_apply(get_single_vif, RHS=RHS)


# ------------------------------------------------------------
# OLSInfluence diagnostic helpers TPU 03/19
# ------------------------------------------------------------
def _check_get_diagnostic_args(lm_grid, diagnostic, do_nobs_loop):
    # type, value checking doesn't run anything, for args see get_diagnostic()

    # types ------------------------------------------------------
    msg = None
    if not isinstance(lm_grid, FitGrid):
        msg = f"lm_grid must be a FitGrid not {type(lm_grid)}"

    if not isinstance(lm_grid.tester, RegressionResultsWrapper):
        msg = f"lm_grid must be fit with fitgrid.lm()"

    if not isinstance(diagnostic, str):
        msg = f"{diagnostic} must be a string"

    if not isinstance(do_nobs_loop, bool):
        msg = f"do_nobs_loop must be True or False"

    if msg is not None:
        raise TypeError(msg)

    # values ------------------------------------------------------
    infl_calc, infl_dtype, index_names = _OLS_INFLUENCE_ATTRS[diagnostic]

    if diagnostic not in _OLS_INFLUENCE_ATTRS:
        msg = f"unknown OLSInfluence attribute {diagnostic}"

    if infl_calc == "nobs_loop" and not do_nobs_loop:
        msg = f"{diagnostic} is slow, set do_nobs_loop=True to calculate"

    if msg is not None:
        raise ValueError(msg)

    if infl_calc is None:
        msg = (
            f"get_diagnostic({diagnostic}), try"
            " lm_grid.get_influence().{diagnostic}"
        )
        raise NotImplementedError(msg)


def _get_diagnostic(lm_grid, diag, do_nobs_loop):
    """grid scraper with a modicum of validation"""

    # modicum of guarding
    _check_get_diagnostic_args(
        lm_grid=lm_grid, diagnostic=diag, do_nobs_loop=do_nobs_loop
    )

    infl_calc, infl_dtype, index_names = _OLS_INFLUENCE_ATTRS[diag]
    attr_df = getattr(lm_grid.get_influence(), diag).copy()

    if not isinstance(attr_df, pd.DataFrame):
        raise TypeError(f"{diag} grid is not a pandas DataFrame")

    actual_type = type(attr_df.iloc[0, 0])
    if actual_type is not infl_dtype:
        raise TypeError(f"gridded {diag} dtype should be {infl_dtype}")

    if not index_names == attr_df.index.names:
        raise TypeError(
            f"_OLS_INFLUENCE_ATTRS thinks {diag} index names"
            f" should be {index_names} not {attr_df.index.names}"
        )

    return attr_df


# ------------------------------------------------------------
# UI wrappers
# ------------------------------------------------------------
def list_diagnostics():
    """Display `statsmodels` diagnostics implemented in fitgrid.utils.lm"""

    fast = [
        f"  get_diagnostic(lm_grid, '{attr}')"
        for attr, spec in _OLS_INFLUENCE_ATTRS.items()
        if spec[0] not in [None, 'nobs_loop']
    ]

    slow = [
        f"  get_diagnostic(lm_grid, '{attr}', do_nobs_loop=True)"
        for attr, spec in _OLS_INFLUENCE_ATTRS.items()
        if spec[0] == 'nobs_loop'
    ]

    not_implemented = [
        (
            f"  {attr}: not implemented for get_diagnostic(), "
            "try querying the grid directly"
        )
        for attr, spec in _OLS_INFLUENCE_ATTRS.items()
        if spec[0] is None
    ]

    print(
        "Fast: These are caclulated quickly with get_diagnostic(),"
        " usable for large data sets\n"
    )
    for usage in fast:
        print(usage)

    print(
        "\nSlow: These are available with get_diagnostic() but"
        " refit a model without each data point. Disabled by"
        " default but can be forced like so\n"
    )
    for usage in slow:
        print(usage)

    print(
        "\nNot implemented:\nThese are not available for get_diagnostic() but"
        "may be available in the fitted grid.\n"
    )
    for usage in not_implemented:
        print(usage)


def get_diagnostic(lm_grid, diagnostic, do_nobs_loop=False):
    """Fetch `statsmodels` diagnostic as a Time  x Channel dataframe

    `statsmodels` implements a variety of data and model diagnostic
    measures. For some, it also computes a version of a recommended
    critical value or :math:`p`-value. Use these at your own risk
    after careful study of the `statsmodels` source code. For details
    visit :sm_docs:`statsmodels.stats.outliers_influence.OLSInfluence.html`

    For a catalog of the measures available for `fitgrid.lm()` run
    this in Python

    .. code-block:: python

       >>>fitgrid.utils.lm.list_diagnostics()

    .. Warning:: Data diagnostics can be very large and very slow, see
       Notes for details.

       * By default **all** values of the diagnostics are computed,
         this dataframe can be pruned with
         :meth:`fitgrid.utils.lm.filter_diagnostic` function.

       * By default slow diagnostics are **not** computed, this can be
         forced by setting `do_nobs_loop=True`.


    Parameters
    ----------
    lm_grid : fitgrid.LMFitGrid
        As returned by :meth:`fitgrid.lm`.

    diagnostic : string
        As implemented in `statsmodels`, e.g., "cooks_distance",
        "dffits_internal", "est_std", "dfbetas".

    do_nobs_loop : bool
        `True` forces slow leave-one-observation-out model refitting.

    Returns
    -------
    diagnostic_df : pandas.DataFrame
        Channels are in columns. Model measures are row indexed by
        time; data measures add an epoch row index; parameter measures add
        a parameter row index.

    sm_1_df : pandas.DataFrame
        The supplemenatary values `statsmodels` returns, or `None`,
        same shape as diagnostic_df.

    Notes
    -----
    * **Size:** `diagnostic_df` values for data measures like
      `cooks_distance` and `hat_matrix_diagonal` are the size of the
      original data plus a row index and for some data measures like
      `dfbetas`, they are the size of the data multiplied by the
      number of regressors in the model.

    * **Speed:** Leave-one-observation-out (LOOO) model refitting takes
      as long as it takes to fit one model multiplied by the number of
      observations. This can be intractable for large
      datasets. Diagnostic measures calculated from the original fit
      like `cooks_distance` and `dffits_internal` are tractable even
      for large data sets.

    Examples
    --------

    .. code-block:: python

       # fake data
       epochs_fg = fitgrid.generate()
       lm_grid = fitgrid.lm(
           epochs_fg,
           LHS=epochs_fg.channels,
           RHS='continuous + categorical',
           parallel=True,
           n_cores=4,
       )

       # data diagnostic, one dataframe with the values
       ess_press, _ = fitgrid.utils.lm.get_diagnostic(
           lm_grid,
           'ess_press'
       )

       # Cook's D dataframe AND the p-values statsmodels computes
       cooks_Ds, sm_pvals = fitgrid.utils.lm.get_diagnostic(
           lm_grid,
           'cooks_distance'
       )

       # this fails because it requires LOOO loop
       dfbetas_df, _  = fitgrid.utils.lm.get_diagnostic(
           lm_grid,
           'dfbetas'
       )

       # this succeeds by forcing LOOO loop calculation
       dfbetas_df, _  = fitgrid.utils.lm.get_diagnostic(
           lm_grid,
           'dfbetas',
           do_nobs_loop=True
       )

    """

    diag_df = _get_diagnostic(lm_grid, diagnostic, do_nobs_loop)

    # special case diagnostic handling is unavoidable b.c. some
    # OLSInflunce methods return 2-ples, most don't. Extract
    # both parts for those that do.
    sm_1_df = None
    if diagnostic in ["cooks_distance", "dffits_internal", "dffits"]:
        assert len(diag_df.index.names) == 3
        assert diag_df.index.names[1] is None  # diagnostic index

        # values returned as 2nd item in diagnostic 2-ple
        sm_1_df = diag_df.loc[pd.IndexSlice[:, 1, :], :].reset_index(
            1, drop=True
        )

        # label the columns
        sm_1_df.columns = pd.MultiIndex.from_product(
            [[f"{diagnostic}_sm_1"], diag_df.columns],
            names=['diagnostic', 'channel'],
        )

        # diagnostic measures
        diag_df = diag_df.loc[pd.IndexSlice[:, 0, :], :].reset_index(
            1, drop=True
        )

    # name unlabeled index from fitgrid
    if len(diag_df.index.names) == 3 and diag_df.index.names[2] is None:
        diag_df.index.names = [
            name if name is not None else f"{diagnostic}_id"
            for name in diag_df.index.names
        ]

    # label the columns
    diag_df.columns = pd.MultiIndex.from_product(
        [[diagnostic], diag_df.columns], names=['diagnostic', 'channel']
    )

    # wide columns == channels diagnostic dataframe
    return diag_df, sm_1_df


def filter_diagnostic(
    diagnostic_df, how, bound_0, bound_1=None, format='long'
):
    """Select a subset of a fitgrid `statsmodels` diagnostic dataframe by value.

    Use this to identify time ponts, epochs, parameters, channels with
    outlying or potentially influential data.

    Parameters
    ----------
    diagnostic_df : pandas.DataFrame
        As returned by :meth:`fitgrid.utils.lm.get_diagnostic`

    how : {'above', 'below', 'inside', 'outside'}
        slice `diagnostic_df` above or below `bound_0` or inside or
        outside the closed interval `(bound_0, bound_1)`.

    bound_0 : scalar or array-like
        `bound_0` is the mandatory boundary for all `how`.  See
        `pandas.DataFrame.gt` and `pandas.DataFrame.lt` documents
        for binary comparisons with dataframes.

    bound_1: scalar or array-like
        `bound_1 is the mandatory upper bound for `how="inside"` and
        `how="outside"`.

    format : {'long', 'wide'}
        The `long` format pivots the channel columns into a row index
        and returns just those times, (epochs, parameters), channels
        that pass the filter. The `wide` format returns `filtered_df`
        with the same shape as `diagnostic_df`, those datapoints that
        pass the filter in their original row, column location, `nans`
        elsewhere.

    Returns
    -------
        selected_df : pandas.DataFrame

    """

    if how in ["above", "below"]:
        try:
            bound_0 > 0
        except Exception as fail:
            fail.args = ("bound_0", *fail.args)
            raise fail

        if bound_1 is not None:
            msg = "bound_1 is ignored with how=above and how=below"
            warnings.warn(msg)
            bound_1 = None

    elif how in ["inside", "outside"]:
        # are bounds comparable, legal
        try:
            bound_1 < bound_0
        except Exception as fail:
            fail.args = ("bound_1, bound_0", *fail.args)
            raise fail

        if np.array(bound_1).__lt__(bound_0):
            msg = "upper bound_1 value(s) less than bound_0"
            raise ValueError(msg)

    else:
        msg = f"how must be above, below, inside, outside"
        raise ValueError(msg)

    # all and only four cases, pandas handles failures
    if how == "above":
        diagnostic_df = diagnostic_df.where(diagnostic_df.gt(bound_0))

    if how == "below":
        diagnostic_df = diagnostic_df.where(diagnostic_df.lt(bound_0))

    if how == "inside":
        diagnostic_df = diagnostic_df.where(
            diagnostic_df.gt(bound_0) & diagnostic_df.lt(bound_1)
        )

    if how == "outside":
        diagnostic_df = diagnostic_df.where(
            diagnostic_df.lt(bound_0) | diagnostic_df.gt(bound_1)
        )

    if format == "long":
        return diagnostic_df.stack(1, dropna=True).sort_index()
    elif format == "wide":
        return diagnostic_df
    else:
        raise ValueError(f"format must be long or wide, not {format}")
